import { Close } from '@mui/icons-material'
import {
  Box,
  CircularProgress,
  Divider,
  FormControl,
  InputLabel,
  MenuItem,
  Select
} from '@mui/material'
import styles from './style.module.scss'
import React, { useEffect, useState } from 'react'

import * as PDFJS from 'pdfjs-dist'
import { ProfileMetadata, User, UserRole } from '../../types'
import {
  PdfFile,
  MouseState,
  PdfPage,
  DrawnField,
  DrawTool
} from '../../types/drawing'
import { truncate } from 'lodash'
import { extractFileExtension, hexToNpub } from '../../utils'
import { toPdfFiles } from '../../utils/pdf.ts'
PDFJS.GlobalWorkerOptions.workerSrc = new URL(
  'pdfjs-dist/build/pdf.worker.min.mjs',
  import.meta.url
).toString()

interface Props {
  selectedFiles: File[]
  users: User[]
  metadata: { [key: string]: ProfileMetadata }
  onDrawFieldsChange: (pdfFiles: PdfFile[]) => void
  selectedTool?: DrawTool
}

export const DrawPDFFields = (props: Props) => {
  const { selectedFiles, selectedTool, onDrawFieldsChange, users } = props

  const [pdfFiles, setPdfFiles] = useState<PdfFile[]>([])
  const [parsingPdf, setParsingPdf] = useState<boolean>(false)

  const [mouseState, setMouseState] = useState<MouseState>({
    clicked: false
  })

  useEffect(() => {
    if (selectedFiles) {
      /**
       * Reads the pdf binary files and converts it's pages to images
       * creates the pdfFiles object and sets to a state
       */
      const parsePdfPages = async () => {
        const pdfFiles: PdfFile[] = await toPdfFiles(selectedFiles)

        setPdfFiles(pdfFiles)
      }

      setParsingPdf(true)

      parsePdfPages().finally(() => {
        setParsingPdf(false)
      })
    }
  }, [selectedFiles])

  useEffect(() => {
    if (pdfFiles) onDrawFieldsChange(pdfFiles)
  }, [onDrawFieldsChange, pdfFiles])

  /**
   * Drawing events
   */
  useEffect(() => {
    // window.addEventListener('mousedown', onMouseDown);
    window.addEventListener('mouseup', onMouseUp)

    return () => {
      // window.removeEventListener('mousedown', onMouseDown);
      window.removeEventListener('mouseup', onMouseUp)
    }
  }, [])

  const refreshPdfFiles = () => {
    setPdfFiles([...pdfFiles])
  }

  /**
   * Fired only when left click and mouse over pdf page
   * Creates new drawnElement and pushes in the array
   * It is re rendered and visible right away
   *
   * @param event Mouse event
   * @param page PdfPage where press happened
   */
  const onMouseDown = (
    event: React.MouseEvent<HTMLDivElement>,
    page: PdfPage
  ) => {
    // Proceed only if left click
    if (event.button !== 0) return

    if (!selectedTool) {
      return
    }

    const { mouseX, mouseY } = getMouseCoordinates(event)

    const newField: DrawnField = {
      left: mouseX,
      top: mouseY,
      width: 0,
      height: 0,
      counterpart: '',
      type: selectedTool.identifier
    }

    page.drawnFields.push(newField)

    setMouseState((prev) => {
      return {
        ...prev,
        clicked: true
      }
    })
  }

  /**
   * Drawing is finished, resets all the variables used to draw
   * @param event Mouse event
   */
  const onMouseUp = () => {
    setMouseState((prev) => {
      return {
        ...prev,
        clicked: false,
        dragging: false,
        resizing: false
      }
    })
  }

  /**
   * After {@link onMouseDown} create an drawing element, this function gets called on every pixel moved
   * which alters the newly created drawing element, resizing it while mouse move
   * @param event Mouse event
   * @param page PdfPage where moving is happening
   */
  const onMouseMove = (
    event: React.MouseEvent<HTMLDivElement>,
    page: PdfPage
  ) => {
    if (mouseState.clicked && selectedTool) {
      const lastElementIndex = page.drawnFields.length - 1

      const lastDrawnField = page.drawnFields[lastElementIndex]

      // Return early if we don't have lastDrawnField
      // Issue noticed in the console when dragging out of bounds
      // to the page below (without releaseing mouse click)
      if (!lastDrawnField) return

      const { mouseX, mouseY } = getMouseCoordinates(event)

      const width = mouseX - lastDrawnField.left
      const height = mouseY - lastDrawnField.top

      lastDrawnField.width = width
      lastDrawnField.height = height

      const currentDrawnFields = page.drawnFields

      currentDrawnFields[lastElementIndex] = lastDrawnField

      refreshPdfFiles()
    }
  }

  /**
   * Fired when event happens on the drawn element which will be moved
   * mouse coordinates relative to drawn element will be stored
   * so when we start moving, offset can be calculated
   * mouseX - offsetX
   * mouseY - offsetY
   *
   * @param event Mouse event
   * @param drawnField Which we are moving
   */
  const onDrawnFieldMouseDown = (event: React.MouseEvent<HTMLDivElement>) => {
    event.stopPropagation()

    // Proceed only if left click
    if (event.button !== 0) return

    const drawingRectangleCoords = getMouseCoordinates(event)

    setMouseState({
      dragging: true,
      clicked: false,
      coordsInWrapper: {
        mouseX: drawingRectangleCoords.mouseX,
        mouseY: drawingRectangleCoords.mouseY
      }
    })
  }

  /**
   * Moves the drawnElement by the mouse position (mouse can grab anywhere on the drawn element)
   * @param event Mouse event
   * @param drawnField which we are moving
   */
  const onDrawnFieldMouseMove = (
    event: React.MouseEvent<HTMLDivElement>,
    drawnField: DrawnField
  ) => {
    if (mouseState.dragging) {
      const { mouseX, mouseY, rect } = getMouseCoordinates(
        event,
        event.currentTarget.parentElement
      )
      const coordsOffset = mouseState.coordsInWrapper

      if (coordsOffset) {
        let left = mouseX - coordsOffset.mouseX
        let top = mouseY - coordsOffset.mouseY

        const rightLimit = rect.width - drawnField.width - 3
        const bottomLimit = rect.height - drawnField.height - 3

        if (left < 0) left = 0
        if (top < 0) top = 0
        if (left > rightLimit) left = rightLimit
        if (top > bottomLimit) top = bottomLimit

        drawnField.left = left
        drawnField.top = top

        refreshPdfFiles()
      }
    }
  }

  /**
   * Fired when clicked on the resize handle, sets the state for a resize action
   * @param event Mouse event
   * @param drawnField which we are resizing
   */
  const onResizeHandleMouseDown = (
    event: React.MouseEvent<HTMLSpanElement>
  ) => {
    // Proceed only if left click
    if (event.button !== 0) return

    event.stopPropagation()

    setMouseState({
      resizing: true
    })
  }

  /**
   * Resizes the drawn element by the mouse position
   * @param event Mouse event
   * @param drawnField which we are resizing
   */
  const onResizeHandleMouseMove = (
    event: React.MouseEvent<HTMLSpanElement>,
    drawnField: DrawnField
  ) => {
    if (mouseState.resizing) {
      const { mouseX, mouseY } = getMouseCoordinates(
        event,
        // currentTarget = span handle
        // 1st parent = drawnField
        // 2nd parent = img
        event.currentTarget.parentElement?.parentElement
      )

      const width = mouseX - drawnField.left
      const height = mouseY - drawnField.top

      drawnField.width = width
      drawnField.height = height

      refreshPdfFiles()
    }
  }

  /**
   * Removes the drawn element using the indexes in the params
   * @param event Mouse event
   * @param pdfFileIndex pdf file index
   * @param pdfPageIndex pdf page index
   * @param drawnFileIndex drawn file index
   */
  const onRemoveHandleMouseDown = (
    event: React.MouseEvent<HTMLSpanElement>,
    pdfFileIndex: number,
    pdfPageIndex: number,
    drawnFileIndex: number
  ) => {
    event.stopPropagation()

    pdfFiles[pdfFileIndex].pages[pdfPageIndex].drawnFields.splice(
      drawnFileIndex,
      1
    )
  }

  /**
   * Used to stop mouse click propagating to the parent elements
   * so select can work properly
   * @param event Mouse event
   */
  const onUserSelectHandleMouseDown = (
    event: React.MouseEvent<HTMLDivElement>
  ) => {
    event.stopPropagation()
  }

  /**
   * Gets the mouse coordinates relative to a element in the `event` param
   * @param event MouseEvent
   * @param customTarget mouse coordinates relative to this element, if not provided
   * event.target will be used
   */
  const getMouseCoordinates = (
    event: React.MouseEvent<HTMLElement>,
    customTarget?: HTMLElement | null
  ) => {
    const target = customTarget ? customTarget : event.currentTarget
    const rect = target.getBoundingClientRect()
    const mouseX = event.clientX - rect.left //x position within the element.
    const mouseY = event.clientY - rect.top //y position within the element.

    return {
      mouseX,
      mouseY,
      rect
    }
  }

  /**
   * Renders the pdf pages and drawing elements
   */
  const getPdfPages = (pdfFile: PdfFile, pdfFileIndex: number) => {
    return (
      <>
        {pdfFile.pages.map((page, pdfPageIndex: number) => {
          return (
            <div
              key={pdfPageIndex}
              className={`${styles.pdfImageWrapper} ${selectedTool ? styles.drawing : ''}`}
            >
              <img
                onMouseMove={(event) => {
                  onMouseMove(event, page)
                }}
                onMouseDown={(event) => {
                  onMouseDown(event, page)
                }}
                draggable="false"
                src={page.image}
              />

              {page.drawnFields.map((drawnField, drawnFieldIndex: number) => {
                return (
                  <div
                    key={drawnFieldIndex}
                    onMouseDown={onDrawnFieldMouseDown}
                    onMouseMove={(event) => {
                      onDrawnFieldMouseMove(event, drawnField)
                    }}
                    className={styles.drawingRectangle}
                    style={{
                      left: `${drawnField.left}px`,
                      top: `${drawnField.top}px`,
                      width: `${drawnField.width}px`,
                      height: `${drawnField.height}px`,
                      pointerEvents: mouseState.clicked ? 'none' : 'all'
                    }}
                  >
                    <span
                      onMouseDown={onResizeHandleMouseDown}
                      onMouseMove={(event) => {
                        onResizeHandleMouseMove(event, drawnField)
                      }}
                      className={styles.resizeHandle}
                    ></span>
                    <span
                      onMouseDown={(event) => {
                        onRemoveHandleMouseDown(
                          event,
                          pdfFileIndex,
                          pdfPageIndex,
                          drawnFieldIndex
                        )
                      }}
                      className={styles.removeHandle}
                    >
                      <Close fontSize="small" />
                    </span>
                    <div
                      onMouseDown={onUserSelectHandleMouseDown}
                      className={styles.userSelect}
                    >
                      <FormControl fullWidth size="small">
                        <InputLabel id="counterparts">Counterpart</InputLabel>
                        <Select
                          value={drawnField.counterpart}
                          onChange={(event) => {
                            drawnField.counterpart = event.target.value
                            refreshPdfFiles()
                          }}
                          labelId="counterparts"
                          label="Counterparts"
                        >
                          {users
                            .filter((u) => u.role === UserRole.signer)
                            .map((user, index) => {
                              let displayValue = truncate(
                                hexToNpub(user.pubkey),
                                {
                                  length: 16
                                }
                              )

                              const metadata = props.metadata[user.pubkey]

                              if (metadata) {
                                displayValue = truncate(
                                  metadata.name ||
                                    metadata.display_name ||
                                    metadata.username,
                                  {
                                    length: 16
                                  }
                                )
                              }

                              return (
                                <MenuItem
                                  key={index}
                                  value={hexToNpub(user.pubkey)}
                                >
                                  {displayValue}
                                </MenuItem>
                              )
                            })}
                        </Select>
                      </FormControl>
                    </div>
                  </div>
                )
              })}
            </div>
          )
        })}
      </>
    )
  }

  if (parsingPdf) {
    return (
      <Box sx={{ width: '100%', textAlign: 'center' }}>
        <CircularProgress />
      </Box>
    )
  }

  if (!pdfFiles.length) {
    return ''
  }

  return (
    <div className={styles.view}>
      {selectedFiles.map((file, i) => {
        const name = file.name
        const extension = extractFileExtension(name)
        const pdfFile = pdfFiles.find((pdf) => pdf.file.name === name)
        return (
          <React.Fragment key={name}>
            <div
              className={`${styles.fileWrapper} ${styles.scrollTarget}`}
              id={`file-${name}`}
            >
              {pdfFile ? (
                getPdfPages(pdfFile, i)
              ) : (
                <div className={styles.otherFile}>
                  This is a {extension} file
                </div>
              )}
            </div>
            {i < selectedFiles.length - 1 && (
              <Divider
                sx={{
                  fontSize: '12px',
                  color: 'rgba(0,0,0,0.15)'
                }}
              >
                File Separator
              </Divider>
            )}
          </React.Fragment>
        )
      })}
    </div>
  )
}
